# Copyright (C) 2012 Aniket Panse <contact@aniketpanse.in
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Aniket Panse <contact@aniketpanse.in> grants Johnny Vestergaard <jkv@unixcluster.dk>
# a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable
# copyright license to reproduce, prepare derivative works of, publicly
# display, publicly perform, sublicense, relicense, and distribute [the] Contributions
# and such derivative works.

import asyncio
import unittest

from heralding.capabilities import http
from heralding.reporting.reporting_relay import ReportingRelay

import http.client as httpclient


class HttpTests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.close()

    def test_connection(self):
        """ Tests if the capability is up, and sending
            HTTP 401 (Unauthorized) headers.
        """
        def http_request():
            client = httpclient.HTTPConnection('127.0.0.1', 8888)
            client.request('GET', '/')
            response = client.getresponse()
            self.assertEqual(response.status, 401)

        options = {'enabled': 'True', 'port': 8888, 'users': {'test': 'test'}}
        http_cap = http.Http(options, self.loop)

        server_coro = asyncio.start_server(http_cap.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        http_task = self.loop.run_in_executor(None, http_request)
        self.loop.run_until_complete(http_task)



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import syslog
import logging

from heralding.reporting.base_logger import BaseLogger

logger = logging.getLogger(__name__)


class SyslogLogger(BaseLogger):
    def __init__(self):
        super().__init__()
        logger.debug('Syslog logger started')

    def handle_log_data(self, data):
        message = "Authentication from {0}:{1}, with username: {2} " \
                  "and password: {3}.".format(data['source_ip'], data['source_port'],
                                              data['username'], data['password'])
        syslog.syslog(syslog.LOG_ALERT, message)



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import zmq
import queue
import logging

import heralding.misc
from heralding.misc.socket_names import SocketNames

logger = logging.getLogger(__name__)


class ReportingRelay:
    _incommingLogQueue = None

    def __init__(self):
        # we are singleton
        assert ReportingRelay._incommingLogQueue is None
        ReportingRelay._incommingLogQueue = queue.Queue(maxsize=10000)

        self.enabled = True

        context = heralding.misc.zmq_context
        self.internalReportingPublisher = context.socket(zmq.PUB)

    @staticmethod
    def queueLogData(data):
        assert ReportingRelay._incommingLogQueue is not None
        ReportingRelay._incommingLogQueue.put(data)

    @staticmethod
    def getQueueSize():
        if ReportingRelay._incommingLogQueue is not None:
            return ReportingRelay._incommingLogQueue.qsize()
        else:
            return 0

    def start(self):
        self.internalReportingPublisher.bind(SocketNames.INTERNAL_REPORTING.value)

        while self.enabled or ReportingRelay.getQueueSize() > 0:
            try:
                data = ReportingRelay._incommingLogQueue.get(timeout=0.5)
                self.internalReportingPublisher.send_pyobj(data)
            except queue.Empty:
                pass

        # None signals 'going down' to listeners
        self.internalReportingPublisher.send_pyobj(None)
        self.internalReportingPublisher.close()

        # None is also used to signal we are all done
        ReportingRelay._incommingLogQueue = None

    def stop(self):
        self.enabled = False



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import uuid
import logging
from datetime import datetime

import heralding.honeypot
from heralding.reporting.reporting_relay import ReportingRelay

logger = logging.getLogger(__name__)


class Session:
    def __init__(self, source_ip, source_port, protocol, users, destination_port=None, destination_ip=None):

        self.id = uuid.uuid4()
        self.source_ip = source_ip
        self.source_port = source_port
        self.protocol = protocol
        self.destination_ip = destination_ip
        self.destination_port = destination_port
        self.timestamp = datetime.utcnow()
        self.login_attempts = 0
        self.session_ended = False

        self.connected = True
        self.authenticated = False
        self.users = users

        # for session specific volatile data (will not get logged)
        self.vdata = {}

        self.last_activity = datetime.utcnow()

    def activity(self):
        self.last_activity = datetime.utcnow()

    def get_number_of_login_attempts(self):
        return self.login_attempts

    def is_connected(self):
        return self.connected

    def add_auth_attempt(self, _type, **kwargs):
        self.login_attempts += 1
        entry = {'timestamp': datetime.utcnow(),
                 'session_id': self.id,
                 'auth_id': uuid.uuid4(),
                 'source_ip': self.source_ip,
                 'source_port': self.source_port,
                 'destination_ip': heralding.honeypot.Honeypot.public_ip,
                 'destination_port': self.destination_port,
                 'protocol': self.protocol,
                 'username': None,
                 'password': None
                 }
        if 'username' in kwargs:
            entry['username'] = kwargs['username']
        if 'password' in kwargs:
            entry['password'] = kwargs['password']

        ReportingRelay.queueLogData(entry)
        self.activity()
        logger.debug('{0} authentication attempt from {1}:{2}. Auth mechanism: {3}. '
                     'Credentials: {4}'.format(self.protocol, self.source_ip,
                                               self.source_port, _type, json.dumps(kwargs)))

    def end_session(self):
        if not self.session_ended:
            self.session_ended = True
            self.connected = False



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import csv
import logging

from heralding.reporting.base_logger import BaseLogger

logger = logging.getLogger(__name__)


class FileLogger(BaseLogger):
    def __init__(self, logFile):
        super().__init__()

        if not os.path.isfile(logFile):
            self.fileHandler = open(logFile, 'w')
        else:
            self.fileHandler = open(logFile, 'a')

        fieldNames = ['timestamp', 'auth_id', 'session_id', 'source_ip', 'source_port', 'destination_ip',
                      'destination_port', 'protocol', 'username', 'password']
        self.csvWriter = csv.DictWriter(self.fileHandler, fieldnames=fieldNames, extrasaction='ignore')

        # empty file, write csv header
        if os.path.getsize(logFile) == 0:
            self.csvWriter.writeheader()
            self.fileHandler.flush()

        logger.info('File logger started, using file: {0}'.format(logFile))

    def loggerStopped(self):
        self.fileHandler.flush()
        self.fileHandler.close()

    def handle_log_data(self, data):
        self.csvWriter.writerow(data)
        # meh
        self.fileHandler.flush()



# We need this in order to reduce the number of third-party modules.
# The main idea is to replace rfile/wfile with reader/writer and to
# add async/await syntax. So, it is http.server.BaseHTTPRequestHandler
# code, but adjusted to work with asyncio in our specific case.

import socket
import http.client

from http import HTTPStatus
from http.server import BaseHTTPRequestHandler

from .aioclient import parse_headers
from heralding.libs.aiobaserequest import AsyncBaseRequestHandler

# It is a hack in order to get the same class as BaseHTTPRequestHandler,
# but which inherits our AsyncBaseRequestHandler.
AsyncHttpHandler = type('AsyncHttpHandler', (AsyncBaseRequestHandler,),
                        dict(BaseHTTPRequestHandler.__dict__))


class AsyncBaseHTTPRequestHandler(AsyncHttpHandler):
    """Asynchronous analogue of http.server.BaseHTTPRequestHandler."""
    async def parse_request(self):
        self.command = None  # set in case of error on the first line
        self.request_version = version = self.default_request_version
        self.close_connection = True
        requestline = str(self.raw_requestline, 'iso-8859-1')
        requestline = requestline.rstrip('\r\n')
        self.requestline = requestline
        words = requestline.split()
        if len(words) == 3:
            command, path, version = words
            if version[:5] != 'HTTP/':
                self.send_error(
                    HTTPStatus.BAD_REQUEST,
                    "Bad request version (%r)" % version)
                return False
            try:
                base_version_number = version.split('/', 1)[1]
                version_number = base_version_number.split(".")
                if len(version_number) != 2:
                    raise ValueError
                version_number = int(version_number[0]), int(version_number[1])
            except (ValueError, IndexError):
                self.send_error(
                    HTTPStatus.BAD_REQUEST,
                    "Bad request version (%r)" % version)
                return False
            if version_number >= (1, 1) and self.protocol_version >= "HTTP/1.1":
                self.close_connection = False
            if version_number >= (2, 0):
                self.send_error(
                    HTTPStatus.HTTP_VERSION_NOT_SUPPORTED,
                    "Invalid HTTP Version (%s)" % base_version_number)
                return False
        elif len(words) == 2:
            command, path = words
            self.close_connection = True
            if command != 'GET':
                self.send_error(
                    HTTPStatus.BAD_REQUEST,
                    "Bad HTTP/0.9 request type (%r)" % command)
                return False
        elif not words:
            return False
        else:
            self.send_error(
                HTTPStatus.BAD_REQUEST,
                "Bad request syntax (%r)" % requestline)
            return False
        self.command, self.path, self.request_version = command, path, version

        # Examine the headers and look for a Connection directive.
        try:
            self.headers = await parse_headers(self.rfile,
                                               _class=self.MessageClass)
        except http.client.LineTooLong:
            self.send_error(
                HTTPStatus.BAD_REQUEST,
                "Line too long")
            return False
        except http.client.HTTPException as err:
            self.send_error(
                HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE,
                "Too many headers",
                str(err)
            )
            return False

        conntype = self.headers.get('Connection', "")
        if conntype.lower() == 'close':
            self.close_connection = True
        elif (conntype.lower() == 'keep-alive' and
                      self.protocol_version >= "HTTP/1.1"):
            self.close_connection = False
        # Examine the headers and look for an Expect directive
        expect = self.headers.get('Expect', "")
        if (expect.lower() == "100-continue" and
                    self.protocol_version >= "HTTP/1.1" and
                    self.request_version >= "HTTP/1.1"):
            if not self.handle_expect_100():
                return False
        return True

    async def handle_one_request(self):
        try:
            self.raw_requestline = await self.rfile.readline()
            if len(self.raw_requestline) > 65536:
                self.requestline = ''
                self.request_version = ''
                self.command = ''
                self.send_error(HTTPStatus.REQUEST_URI_TOO_LONG)
                return
            if not self.raw_requestline:
                self.close_connection = True
                return
            if not await self.parse_request():
                # An error code has been sent, just exit
                return
            mname = 'do_' + self.command
            if not hasattr(self, mname):
                self.send_error(
                    HTTPStatus.NOT_IMPLEMENTED,
                    "Unsupported method (%r)" % self.command)
                return
            method = getattr(self, mname)
            method()
        except socket.timeout as e:
            #a read or a write timed out.  Discard this connection
            self.log_error("Request timed out: %r", e)
            self.close_connection = True
            return

    async def handle(self):
        self.close_connection = True

        await self.handle_one_request()
        while not self.close_connection:
            await self.handle_one_request()



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import zmq
import logging

import heralding.misc
from heralding.misc.socket_names import SocketNames

logger = logging.getLogger(__name__)


class BaseLogger:
    def __init__(self):
        self.enabled = True

    def start(self):
        context = heralding.misc.zmq_context

        internal_reporting_socket = context.socket(zmq.SUB)
        internal_reporting_socket.connect(SocketNames.INTERNAL_REPORTING.value)
        internal_reporting_socket.setsockopt(zmq.SUBSCRIBE, b'')

        poller = zmq.Poller()
        poller.register(internal_reporting_socket, zmq.POLLIN)
        while self.enabled:
            socks = dict(poller.poll(500))
            if internal_reporting_socket in socks and socks[internal_reporting_socket] == zmq.POLLIN:
                data = internal_reporting_socket.recv_pyobj()
                # if None is received, this means that ReportingRelay is going down
                if not data:
                    self.stop()
                else:
                    assert isinstance(data, dict)
                    self.handle_log_data(data)
        internal_reporting_socket.close()
        # at this point we know no more data will arrive.
        self.loggerStopped()

    def stop(self):
        self.enabled = False

    def handle_log_data(self, data):
        # should be handled in child class
        raise NotImplementedError("Please implement this method")

    # called after we are sure no more data is received
    # override this to close filesockets, etc.
    def loggerStopped(self):
        pass



# We need this in order to reduce the number of third-party modules.
# It is a part of code from socketserver, that adjusted to work with
# asyncio in our specific case.


class AsyncBaseRequestHandler:
    """Asynchronous analogue of socketserver.BaseRequestHandler."""
    def __init__(self, reader, writer, client_address):
        self.rfile = reader
        self.wfile = writer
        self.client_address = client_address

    async def run(self):
        self.setup()
        try:
            await self.handle()
        finally:
            self.finish()

    def setup(self):
        pass

    async def handle(self):
        pass

    def finish(self):
        pass



import os
import glob

# Detect all modules
for fullname in glob.glob(os.path.dirname(__file__) + "/*.py"):
    name = os.path.basename(fullname)
    # __init__ and handlerbase are not capabilities, so ignore them
    if name[:-3] != "__init__" or name[:-3] != "handlerbase":
        __import__("heralding.capabilities." + name[:-3])



# Copyright (C) 2017 Roman Samoilenko <ttahabatt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import functools

from heralding.capabilities.handlerbase import HandlerBase

import asyncssh
from Crypto.PublicKey import RSA

logger = logging.getLogger(__name__)


class SSH(asyncssh.SSHServer, HandlerBase):
    connections_list = []

    def __init__(self, options, loop):
        asyncssh.SSHServer.__init__(self)
        HandlerBase.__init__(self, options, loop)

    def connection_made(self, conn):
        SSH.connections_list.append(conn)
        self.address = conn.get_extra_info('peername')
        self.handle_connection()
        logger.debug('SSH connection received from %s.' % conn.get_extra_info('peername')[0])

    def connection_lost(self, exc):
        self.close_session(self.session)
        if exc:
            logger.debug('SSH connection error: ' + str(exc))
        else:
            logger.debug('SSH connection closed.')

    def begin_auth(self, username):
        return True

    def password_auth_supported(self):
        return True

    def validate_password(self, username, password):
        self.session.add_auth_attempt('plaintext', username=username, password=password)
        return False

    def handle_connection(self):
        if HandlerBase.global_sessions > HandlerBase.MAX_GLOBAL_SESSIONS:
            protocol = self.__class__.__name__.lower()
            logger.warning(
                'Got {0} session on port {1} from {2}:{3}, but not handling it because the global session limit has '
                'been reached'.format(protocol, self.port, *self.address))
        else:
            self.session = self.create_session(self.address)

    @staticmethod
    def change_server_banner(banner):
        """_send_version code was copied from asyncssh.connection in order to change
        internal local variable 'version', providing custom banner."""

        @functools.wraps(asyncssh.connection.SSHConnection._send_version)
        def _send_version(self):
            """Start the SSH handshake"""

            version = bytes(banner, 'utf-8')

            if self.is_client():
                self._client_version = version
                self._extra.update(client_version=version.decode('ascii'))
            else:
                self._server_version = version
                self._extra.update(server_version=version.decode('ascii'))

            self._send(version + b'\r\n')

        asyncssh.connection.SSHConnection._send_version = _send_version

    @staticmethod
    def generate_ssh_key(ssh_key_file):
        if not os.path.isfile(ssh_key_file):
            with open(ssh_key_file, 'w') as _file:
                rsa_key = RSA.generate(2048)
                priv_key_text = str(rsa_key.exportKey('PEM', pkcs=1), 'utf-8')
                _file.write(priv_key_text)



# Copyright (C) 2010-2013 Mark Schloesser <ms@mwcollect.org
# This file is part of hpfeeds - https://github.com/rep/hpfeeds
# See the file 'LICENSE' for copying permission.

import sys
import struct
import socket
import hashlib
import logging
import time
import threading
import ssl

logger = logging.getLogger('pyhpfeeds')

OP_ERROR	= 0
OP_INFO		= 1
OP_AUTH		= 2
OP_PUBLISH	= 3
OP_SUBSCRIBE	= 4
BUFSIZ = 16384

__all__ = ["new", "FeedException"]

def msghdr(op, data):
	return struct.pack('!iB', 5+len(data), op) + data
def msgpublish(ident, chan, data):
#	if isinstance(data, str):
#		data = data.encode('latin1')
	return msghdr(OP_PUBLISH, struct.pack('!B', len(ident)) + ident + struct.pack('!B', len(chan)) + chan + data)
def msgsubscribe(ident, chan):
	return msghdr(OP_SUBSCRIBE, struct.pack('!B', len(ident)) + ident + chan)
def msgauth(rand, ident, secret):
	hash = hashlib.sha1(rand+secret).digest()
	return msghdr(OP_AUTH, struct.pack('!B', len(ident)) + ident + hash)

class FeedUnpack(object):
	def __init__(self):
		self.buf = bytearray()
	def __iter__(self):
		return self
	def __next__(self):
		return self.unpack()
	def feed(self, data):
		self.buf.extend(data)
	def unpack(self):
		if len(self.buf) < 5:
			raise StopIteration('No message.')

		ml, opcode = struct.unpack('!iB', buffer(self.buf,0,5))
		if len(self.buf) < ml:
			raise StopIteration('No message.')

		data = bytearray(buffer(self.buf, 5, ml-5))
		del self.buf[:ml]
		return opcode, data

class FeedException(Exception):
	pass
class Disconnect(Exception):
	pass

class HPC(object):
	def __init__(self, host, port, ident, secret, timeout=3, reconnect=True, sleepwait=20):
		self.host, self.port = host, port
		self.ident, self.secret = ident, secret
		self.timeout = timeout
		self.reconnect = reconnect
		self.sleepwait = sleepwait
		self.brokername = 'unknown'
		self.connected = False
		self.stopped = False
		self.s = None
		self.connecting_lock = threading.Lock()
		self.subscriptions = set()
		self.unpacker = FeedUnpack()

		self.tryconnect()

	def makesocket(self, addr_family):
		return socket.socket(addr_family, socket.SOCK_STREAM)

	def recv(self):
		try:
			d = self.s.recv(BUFSIZ)
		except socket.timeout:
			return ""
		except socket.error as e:
			logger.warn("Socket error: %s", e)
			raise Disconnect()

		if not d: raise Disconnect()
		return d

	def send(self, data):
		try:
			self.s.sendall(data)
		except socket.timeout:
			logger.warn("Timeout while sending - disconnect.")
			raise Disconnect()
		except socket.error as e:
			logger.warn("Socket error: %s", e)
			raise Disconnect()

		return True

	def tryconnect(self):
		with self.connecting_lock:
			if not self.connected:
				while True:
					try:
						self.connect()
						break
					except socket.error as e:
						logger.warn('Socket error while connecting: {0}'.format(e))
						time.sleep(self.sleepwait)
					except FeedException as e:
						logger.warn('FeedException while connecting: {0}'.format(e))
						time.sleep(self.sleepwait)
					except Disconnect as e:
						logger.warn('Disconnect while connecting.')
						time.sleep(self.sleepwait)

	def connect(self):
		self.close_old()

		logger.info('connecting to {0}:{1}'.format(self.host, self.port))

		# Try other resolved addresses (IPv4 or IPv6) if failed.
		ainfos = socket.getaddrinfo(self.host, 1, socket.AF_UNSPEC, socket.SOCK_STREAM)
		for ainfo in ainfos:
			addr_family = ainfo[0]
			addr = ainfo[4][0]
			try:
				self.s = self.makesocket(addr_family)
				self.s.settimeout(self.timeout)
				self.s.connect((addr, self.port))
			except:
				import traceback
				traceback.print_exc()
				#print 'Could not connect to broker. %s[%s]' % (self.host, addr)
				continue
			else:
				self.connected = True
				break

		if self.connected == False:
			raise FeedException('Could not connect to broker [%s].' % (self.host))

		try: d = self.s.recv(BUFSIZ)
		except socket.timeout: raise FeedException('Connection receive timeout.')

		self.unpacker.feed(d)
		for opcode, data in self.unpacker:
			if opcode == OP_INFO:
				rest = buffer(data, 0)
				name, rest = rest[1:1+ord(rest[0])], buffer(rest, 1+ord(rest[0]))
				rand = str(rest)

				logger.debug('info message name: {0}, rand: {1}'.format(name, repr(rand)))
				self.brokername = name

				self.send(msgauth(rand, self.ident, self.secret))
				break
			else:
				raise FeedException('Expected info message at this point.')

		self.s.settimeout(None)
		self.s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

		if sys.platform in ('linux2', ):
			self.s.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 10)    

	def run(self, message_callback, error_callback):
		while not self.stopped:
			self._subscribe()
			while self.connected:
				try:
					d = self.recv()
					self.unpacker.feed(d)

					for opcode, data in self.unpacker:
						if opcode == OP_PUBLISH:
							rest = buffer(data, 0)
							ident, rest = rest[1:1+ord(rest[0])], buffer(rest, 1+ord(rest[0]))
							chan, content = rest[1:1+ord(rest[0])], buffer(rest, 1+ord(rest[0]))

							message_callback(str(ident), str(chan), content)
						elif opcode == OP_ERROR:
							error_callback(data)

				except Disconnect:
					self.connected = False
					logger.info('Disconnected from broker.')
					break

				# end run loops if stopped
				if self.stopped: break

			if not self.stopped and self.reconnect:
				# connect again if disconnected
				self.tryconnect()

		logger.info('Stopped, exiting run loop.')

	def wait(self, timeout=1):
		self.s.settimeout(timeout)

		try:
			d = self.recv()
			if not d: return None

			self.unpacker.feed(d)
			for opcode, data in self.unpacker:
				if opcode == OP_ERROR:
					return data
		except Disconnect:
			pass

		return None

	def close_old(self):
		if self.s:
			try: self.s.close()
			except: pass

	def subscribe(self, chaninfo):
		if type(chaninfo) == str:
			chaninfo = [chaninfo,]
		for c in chaninfo:
			self.subscriptions.add(c)

	def _subscribe(self):
		for c in self.subscriptions:
			try:
				logger.debug('Sending subscription for {0}.'.format(c))
				self.send(msgsubscribe(self.ident, c))
			except Disconnect:
				self.connected = False
				logger.info('Disconnected from broker (in subscribe).')
				if not self.reconnect: raise
				break

	def publish(self, chaninfo, data):
		if type(chaninfo) == str:
			chaninfo = [chaninfo,]
		for c in chaninfo:
			try:
				self.send(msgpublish(self.ident, c, data))
			except Disconnect:
				self.connected = False
				logger.info('Disconnected from broker (in publish).')
				if self.reconnect:
					self.tryconnect()
				else:
					raise

	def stop(self):
		self.stopped = True

	def close(self):
		try: self.s.close()
		except: logger.debug('Socket exception when closing (ignored though).')


class HPC_SSL(HPC):
	def __init__(self, *args, **kwargs):
		self.certfile = kwargs.pop("certfile", None)
		HPC.__init__(self, *args, **kwargs)

	def makesocket(self, addr_family):
		s = socket.socket(addr_family, socket.SOCK_STREAM)
		return ssl.wrap_socket(s, ca_certs=self.certfile, ssl_version=3, cert_reqs=2)


def new(host=None, port=10000, ident=None, secret=None, timeout=3, reconnect=True, sleepwait=20, certfile=None):
	if certfile:
		return HPC_SSL(host, port, ident, secret, timeout, reconnect, certfile=certfile)
	return HPC(host, port, ident, secret, timeout, reconnect)


# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging

from heralding.capabilities.handlerbase import HandlerBase
from heralding.capabilities.imap import Imap

logger = logging.getLogger(__name__)


class Imaps(Imap, HandlerBase):
    """
    This class will get wrapped in SSL. This is possible because we by convention  wrap
    all capabilities that ends with the letter 's' in SSL."""



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import unittest

import ftplib
from ftplib import FTP

from heralding.capabilities import ftp
from heralding.reporting.reporting_relay import ReportingRelay


class FtpTests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.close()

    def test_login(self):
        """Testing different login combinations"""

        def ftp_login():
            ftp_client = FTP()
            ftp_client.connect('127.0.0.1', 8888, 1)
            # expect perm exception
            try:
                ftp_client.login('james', 'bond')
                _ = ftp_client.getresp()  # NOQA
            except ftplib.error_perm:
                ftp_client.quit()

        options = {'enabled': 'True', 'port': 0, 'banner': 'Test Banner', 'users': {'test': 'test'},
                   'protocol_specific_data': {'max_attempts': 3, 'banner': 'test banner', 'syst_type': 'Test Type'}}

        ftp_capability = ftp.ftp(options, self.loop)

        server_coro = asyncio.start_server(ftp_capability.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        ftp_task = self.loop.run_in_executor(None, ftp_login)
        self.loop.run_until_complete(ftp_task)



from ez_setup import use_setuptools
use_setuptools()  # NOQA
from setuptools import setup, find_packages
from heralding import version as heralding_version

setup(
    name='Heralding',
    version=heralding_version,
    packages=find_packages(exclude=['bin', 'docs', '*.pyc']),
    scripts=['bin/heralding'],
    url='https://github.com/johnnykv/heralding',
    license='GPL 3',
    author='Johnny Vestergaard',
    author_email='jkv@unixcluster.dk',
    include_package_data=True,
    long_description=open('README.rst').read(),
    description='Credentials catching honeypot.',
    test_suite='nose.collector',
    install_requires=open('requirements.txt').read(),
    tests_require=open('requirements-test.txt').read(),
    package_data={
        "heralding": ["heralding.yml"],
    },
)



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import unittest

from heralding.capabilities.ssh import SSH
from heralding.reporting.reporting_relay import ReportingRelay

import asyncssh


class SshTests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.close()

    def test_basic_login(self):
        async def run_client():
            async with asyncssh.connect('localhost', port=8888,
                                        username='johnny', password='secretpw',
                                        known_hosts=None, loop=self.loop) as _:
                pass

        ssh_key_file = 'ssh.key'
        SSH.generate_ssh_key(ssh_key_file)

        options = {'enabled': 'True', 'port': 8888}
        server_coro = asyncssh.create_server(lambda: SSH(options, self.loop), '0.0.0.0', 8888,
                                             server_host_keys=['ssh.key'], loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        try:
            self.loop.run_until_complete(run_client())
        except asyncssh.Error:
            pass



# Copyright (C) 2013 Aniket Panse <contact@aniketpanse.in>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Parts of this code are from secure-smtpd (https://github.com/bcoe/secure-smtpd)

# Aniket Panse <contact@aniketpanse.in> grants Johnny Vestergaard <jkv@unixcluster.dk>
# a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable
# copyright license to reproduce, prepare derivative works of, publicly
# display, publicly perform, sublicense, relicense, and distribute [the] Contributions
# and such derivative works.

import time
import random
import base64
import socket
import asyncio
import logging

from aiosmtpd.smtp import SMTP, MISSING, syntax

from heralding.capabilities.handlerbase import HandlerBase

log = logging.getLogger(__name__)


class SMTPHandler(SMTP):
    fqdn = ''

    def __init__(self, reader, writer, session, options, loop):
        self.banner = options['protocol_specific_data']['banner']
        super().__init__(None, hostname=self.banner, loop=loop)
        # Reset standard banner.
        self.__ident__ = ""
        self._reader = reader
        self._writer = writer
        self.transport = writer

        self._set_rset_state()
        self.session = session
        self.session.peer = self.transport.get_extra_info('peername')
        self.session.extended_smtp = None
        self.session.host_name = None

    async def push(self, status):
        response = bytes(
            status + '\r\n', 'utf-8' if self.enable_SMTPUTF8 else 'ascii')
        self._writer.write(response)
        log.debug(response)
        try:
            await self._writer.drain()
        except ConnectionResetError:
            self.stop()

    @syntax('EHLO hostname')
    async def smtp_EHLO(self, hostname):
        if not hostname:
            await self.push('501 Syntax: EHLO hostname')
            return
        self._set_rset_state()
        await self.push('250-{0} Hello {1}'.format(self.hostname, hostname))
        await self.push('250-AUTH PLAIN LOGIN CRAM-MD5')
        await self.push('250 EHLO')

    @syntax("AUTH mechanism [initial-response]")
    async def smtp_AUTH(self, arg):
        if not arg:
            await self.push('500 Not enough values')
            return
        args = arg.split()
        if len(args) > 2:
            await self.push('500 Too many values')
            return
        mechanism = args[0]
        if mechanism == 'PLAIN':
            if len(args) == 1:
                await self.push('334 ')  # wait for client login/password
                line = await self.readline()
                if not line:
                    return
                blob = line.strip()
            else:
                blob = args[1].encode()

            try:
                loginpassword = base64.b64decode(blob)
            except Exception:
                await self.push("501 Can't decode base64")
                return
            try:
                _, login, password = loginpassword.split(b"\x00")
            except ValueError:  # not enough args
                await self.push("500 Can't split auth value")
                return
            self.session.add_auth_attempt('PLAIN', username=str(login, 'utf-8'),
                                          password=str(password, 'utf-8'))
        elif mechanism == 'LOGIN':
            if len(args) > 1:
                username = str(base64.b64decode(args[1]), 'utf-8')
                await self.push('334 ' + str(base64.b64encode(b'Password:'), 'utf-8'))

                password_bytes = await self.readline()
                if not password_bytes:
                    return
                password = str(base64.b64decode(password_bytes), 'utf-8')
                self.session.add_auth_attempt('LOGIN', username=username, password=password)
            else:
                await self.push('334 ' + str(base64.b64encode(b'Username:'), 'utf-8'))

                username_bytes = await self.readline()
                if not username_bytes:
                    return

                await self.push('334 ' + str(base64.b64encode(b'Password:'), 'utf-8'))

                password_bytes = await self.readline()
                if not password_bytes:
                    return
                self.session.add_auth_attempt('LOGIN', username=str(base64.b64decode(username_bytes), 'utf-8'),
                                              password=str(base64.b64decode(password_bytes), 'utf-8'))
        elif mechanism == 'CRAM-MD5':
            r = random.randint(5000, 20000)
            t = int(time.time())

            # challenge is of the form '<24609.1047914046@awesome.host.com>'
            sent_cram_challenge = "<" + str(r) + "." + str(t) + "@" + SMTPHandler.fqdn + ">"
            cram_challenge_bytes = bytes(sent_cram_challenge, 'utf-8')
            await self.push("334 " + str(base64.b64encode(cram_challenge_bytes), 'utf-8'))

            credentials_bytes = await self.readline()
            if not credentials_bytes:
                return
            credentials = str(base64.b64decode(credentials_bytes), 'utf-8')
            if sent_cram_challenge is None or ' ' not in credentials:
                await self.push('451 Internal confusion')
                return
            username, digest = credentials.split()
            self.session.add_auth_attempt('cram_md5', username=username,
                                          digest=digest, challenge=sent_cram_challenge)
            await self.push('535 authentication failed')
        else:
            await self.push('500 incorrect AUTH mechanism')
            return
        status = '535 authentication failed'
        await self.push(status)

    @syntax('QUIT')
    async def smtp_QUIT(self, arg):
        if arg:
            await self.push('501 Syntax: QUIT')
        else:
            status = await self._call_handler_hook('QUIT')
            await self.push('221 Bye' if status is MISSING else status)
            self.stop()

    async def readline(self):
        line = b''
        try:
            line = await self._reader.readline()
        except ConnectionResetError:
            self.stop()
        else:
            return line

    def stop(self):
        self.transport.close()
        self.transport = None


class smtp(HandlerBase):
    def __init__(self, options, loop):
        super().__init__(options, loop)
        self.loop = loop
        self._options = options

    async def execute_capability(self, reader, writer, session):
        fqdn_task = asyncio.ensure_future(self.setfqdn(), loop=self.loop)

        smtp_cap = SMTPHandler(reader, writer, session, self._options, self.loop)
        smtp_task = asyncio.ensure_future(smtp_cap._handle_client(), loop=self.loop)

        await smtp_task

        fqdn_task.cancel()
        try:
            await fqdn_task
        except asyncio.CancelledError:
            pass

    async def setfqdn(self):
        if 'fqdn' in self._options['protocol_specific_data'] and self._options['protocol_specific_data']['fqdn']:
            SMTPHandler.fqdn = self._options['protocol_specific_data']['fqdn']
        else:
            while True:
                fqdn = await self.loop.run_in_executor(None, socket.getfqdn)
                SMTPHandler.fqdn = fqdn
                await asyncio.sleep(1800, loop=self.loop)



# -*- coding: utf-8 -*-
# Copyright (C) 2017 Roman Samoilenko <ttahabatt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import imaplib
import asyncio
import unittest

from heralding.capabilities.imap import Imap
from heralding.reporting.reporting_relay import ReportingRelay


class ImapTests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.close()

    def test_LOGIN(self):
        """Testing different login combinations using simple login auth mechanism."""

        def imap_login():
            login_sequences = [
                ('kajoj_admin', 'thebestpassword'),
                ('\"kajoj_admin\"', 'the best password')
            ]

            imap_obj = imaplib.IMAP4('127.0.0.1', port=8888)
            for sequence in login_sequences:
                with self.assertRaises(imaplib.IMAP4.error) as error:
                    imap_obj.login(sequence[0], sequence[1])
                imap_exception = error.exception
                self.assertEqual(imap_exception.args[0], b'Authentication failed')
            imap_obj.logout()

        options = {'enabled': 'True', 'port': 143, 'timeout': 30,
                   'protocol_specific_data': {'max_attempts': 3,
                                              'banner': '* OK IMAP4rev1 Server Ready'}}
        capability = Imap(options, self.loop)
        server_coro = asyncio.start_server(capability.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        imap_task = self.loop.run_in_executor(None, imap_login)
        self.loop.run_until_complete(imap_task)

    def test_AUTHENTICATE_PLAIN(self):
        """Testing different login combinations using plain auth mechanism."""

        def imap_authenticate():
            # imaplib in Python 3.5.3 and higher returns str representation of auth failure
            # But imaplib in Python 3.5.2 and lower returns bytes.
            # This is a sad hack to get around this problem.
            pyversion = sys.version_info[:3]
            if pyversion < (3, 5, 3):
                auth_failure_msg = b'Authentication failed'
            else:
                auth_failure_msg = 'Authentication failed'
            login_sequences = [
                ('\0kajoj_admin\0thebestpassword', auth_failure_msg),
                ('\0пайтон\0наилучшийпароль', auth_failure_msg),
                ('kajoj_admin\0the best password', 'AUTHENTICATE command error: BAD [b\'invalid command\']')
            ]

            imap_obj = imaplib.IMAP4('127.0.0.1', port=8888)
            for sequence in login_sequences:
                with self.assertRaises(imaplib.IMAP4.error) as error:
                    imap_obj.authenticate('PLAIN', lambda x: sequence[0])
                imap_exception = error.exception
                self.assertEqual(imap_exception.args[0], sequence[1])
            imap_obj.logout()

        options = {'enabled': 'True', 'port': 143, 'timeout': 30,
                   'protocol_specific_data': {'max_attempts': 3,
                                              'banner': '* OK IMAP4rev1 Server Ready'}}
        capability = Imap(options, self.loop)

        server_coro = asyncio.start_server(capability.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        imap_task = self.loop.run_in_executor(None, imap_authenticate)
        self.loop.run_until_complete(imap_task)


if __name__ == '__main__':
    unittest.main()
